/***************************************************************************
 *   Copyright (C) 2007 by Vincenzo Forchi`,,,   *
 *   vincenzo.forchi@elettra.trieste.it   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#include "interlock.h"
#include "Sound.h"
#include <qtango.h>
#include <device_factory.h>
#include <QtSql>
#include <QScrollBar>
#include "mysqlmodel.h"
#include <TLog>
#include <QMessageBox>
#include <elettracolors.h>

Interlock::Interlock(QWidget *parent) : QWidget(parent)
{
    QStringList args = qApp->arguments();
    if(args.size() < 5)
    {
        QMessageBox::information(this, "Usage",
                                 QString("%1 f/interlock/plc_mod_ll f/interlock/plc_int_ll.01"
                                         "\nf/interlock/plc_int_ll.02 srv-padres-srf:20000/srv-tango-srf/interlock/plc_int_ll.03").
                                 arg(args.first()));
        exit(EXIT_FAILURE);
    }

    if(args.contains("--sound-file-name") && args.size() > args.indexOf("--sound-file-name"))
    {
        Sound *sound = new Sound(this, args.at(args.indexOf("--sound-file-name") + 1));
    }
    else
    {
        printf("\e[1;35m - if you want to enable sounds for alarms, provide as second argument the name of the sound file\e[0m\n");
        ui.pbTestSound->setHidden(true);
        ui.line->setHidden(true);
    }

    modDeviceName = QApplication::arguments()[1];
    vacPsDeviceName = QApplication::arguments()[2];

    Config::instance()->setOverrideValuesAttributePropertyName("labels");

    ui.setupUi(this);
    setupDb();
    configureUi();
    createConnections();
    QTimer::singleShot(50, this, SLOT(connectTango()));
}

void Interlock::setupDb()
{
    /* connect to db */
    db = QSqlDatabase::addDatabase("QMYSQL");
    db.setHostName("srv-db-srf-02");
    db.setDatabaseName("fermi_interlock");
    db.setUserName("firo");
    db.setPassword("firo_password");
    if (!db.open())
        QMessageBox::critical(0, "Error", QString("Can't establish connection with db \"%1\" on host \"%2\"").arg(db.databaseName()).arg(db.hostName()));

    queryType = QueryConfiguration::LastDay;
    modulatorModel = new MySqlModel();
    psvacModel = new MySqlModel();
    ui.treeLogs->setModel(modulatorModel);
    ui.treeLogsPSVAC->setModel(psvacModel);

    dbTimer = new QTimer(this);
    connect(dbTimer, SIGNAL(timeout()), this, SLOT(readFromDb()));
}

void Interlock::createConnections()
{
    connect(ui.treeAlarms, SIGNAL(acknowledgeSelected()), this, SLOT(acknowledgeSelected()));
    connect(ui.pushRefresh, SIGNAL(clicked()), this, SLOT(readFromDb()));
    connect(ui.pushRefreshPSVacuum, SIGNAL(clicked()), this, SLOT(readFromDb()));
    connect(ui.pushQueryConfiguration, SIGNAL(clicked()), this, SLOT(queryConfiguration()));
    connect(ui.pushQueryConfigurationPSVAC, SIGNAL(clicked()), this, SLOT(queryConfiguration()));
    connect(ui.comboAck, SIGNAL(currentIndexChanged(int)), this, SLOT(changeAckTarget(int)));
    connect(ui.tabWidget, SIGNAL(currentChanged(int)), this, SLOT(tabChanged(int)));

    /* 1 tab: mod filter */
    connect(ui.pbClearModFilter, SIGNAL(clicked()), ui.leFilterMod, SLOT(clear()));
    connect(ui.leFilterMod, SIGNAL(textChanged(QString)), ui.treeAlarms, SLOT(filter(QString)));

    /* 2 tab: mod history */
    connect(ui.pbClearModHist, SIGNAL(clicked()), ui.leFilterModHist, SLOT(clear()));
    connect(ui.leFilterModHist, SIGNAL(textChanged(QString)), ui.treeLogs, SLOT(filter(QString)));

    /* 3 tab: ps vac linac */
    connect(ui.pbClearFilterVacLinac, SIGNAL(clicked()), ui.leFilterPsVacLinac, SLOT(clear()));
    connect(ui.leFilterPsVacLinac, SIGNAL(textChanged(QString)), ui.treeAlarmsVPS, SLOT(filter(QString)));

    /* 4 tab: ps vacuum uh */
    connect(ui.pbClearFilterPSVacUH, SIGNAL(clicked()), ui.leFilterPsVacUH, SLOT(clear()));
    connect(ui.leFilterPsVacUH, SIGNAL(textChanged(QString)), ui.treeAlarmsVUH, SLOT(filter(QString)));

    /* 5 tab ps vacuum padres */
    connect(ui.pbClearFilterPsVacPadres, SIGNAL(clicked()), ui.leFilterPsVacPadres, SLOT(clear()));
    connect(ui.leFilterPsVacPadres, SIGNAL(textChanged(QString)), ui.treeAlarmsPadres, SLOT(filter(QString)));

    /* 6. tab ps and vacuum history */
    connect(ui.pbClearFilterPsVacHistory, SIGNAL(clicked()), ui.leFilterPsVacHistory, SLOT(clear()));
    connect(ui.leFilterPsVacHistory, SIGNAL(textChanged(QString)), ui.treeLogsPSVAC, SLOT(filter(QString)));

    /* sound */
    Sound *sound = findChild<Sound *>();
    if(sound)
    {
        connect(ui.treeAlarms,          SIGNAL(newAlarmArrived()), sound, SLOT(play()));
        connect(ui.treeAlarmsPadres,    SIGNAL(newAlarmArrived()), sound, SLOT(play()));
        connect(ui.treeAlarmsVPS,       SIGNAL(newAlarmArrived()), sound, SLOT(play()));
        connect(ui.treeAlarmsVUH,       SIGNAL(newAlarmArrived()), sound, SLOT(play()));
        connect(ui.pbTestSound, SIGNAL(clicked()), sound, SLOT(play()));
    }
    connect(ui.treeAlarms,          SIGNAL(newAlarmArrived()), this, SLOT(newAlarmArrived()));
    connect(ui.treeAlarmsPadres,    SIGNAL(newAlarmArrived()), this, SLOT(newAlarmArrived()));
    connect(ui.treeAlarmsVPS,       SIGNAL(newAlarmArrived()), this, SLOT(newAlarmArrived()));
    connect(ui.treeAlarmsVUH,       SIGNAL(newAlarmArrived()), this, SLOT(newAlarmArrived()));
}

void Interlock::configureUi()
{
    QStringList headers;
    headers << "Date Time" << "Name" << "Description";
    foreach(TTreeWidget *ttw, findChildren<TTreeWidget *>(QRegExp("treeAlarms.*")))
        ttw->setHeaderLabels(headers);
    QStringList mods;
    for(int i = 0; i <= 15; i++)
        mods << QString("mod%1").arg(i, 2, 10, QChar('0'));
    mods.push_front("General");
    ui.comboAck->insertItems(0, mods);
}

Interlock::~Interlock()
{
}

void Interlock::connectTango()
{
    /* mod */
    ui.treeAlarms->setPeriod(1000);
    ui.treeAlarms->setSource(modDeviceName + "/Db_alarm");
    ui.tAck->setTargets(modDeviceName + QString("->Acknowledge(%1)").arg(ui.comboAck->itemData(0).toInt()));
    /* vacuum and power supply */
    ui.treeAlarmsVPS->setPeriod(1000);
    ui.treeAlarmsVPS->setSource(vacPsDeviceName + "/Db_alarm");
    /* Ps and vacuum UH */
    ui.treeAlarmsVUH->setSource("$3/Db_alarm");
    /* padres */
    ui.treeAlarmsPadres->setSource("$4/Db_alarm");
}

void Interlock::changeAckTarget(int idx)
{
    /* item list (idx == 0) starts with "general".
   * Then modulators follow, from mod00 (idx == 1) to mod12.
   * So the correct argin is obtained subtracting 1 to the combo box index.
   */
    int argin = idx - 1;
    ui.tAck->setTargets(modDeviceName + QString("->Acknowledge(%1)").arg(argin));
}

void Interlock::newAlarmArrived()
{
    /* get sender */
    QString oName = sender()->objectName();
    int tabIndex = -1;
    if(oName == "treeAlarms")
        tabIndex = 0;
    else if(oName == "treeAlarmsVPS")
        tabIndex = 2;
    else if(oName == "treeAlarmsVUH")
        tabIndex = 3;
    else if(oName == "treeAlarmsPadres")
        tabIndex = 4;

    ui.tabWidget->setTabIcon(tabIndex, QIcon(":icons/alarm.png"));
}

/* icon removal is made inside custom TabBar.
 * See tabwidget.h
 */
void Interlock::tabChanged(int idx)
{
    if(idx == ModHistory || idx == PsVacHistory)
    {
        dbTimer->start(10000);
        this->readFromDb();
    }
    else
    {
        dbTimer->stop();
    }
}

void Interlock::acknowledgeSelected()
{
    /// //////////////////
    printf("\e[1;33m - da implementare - \e[0m\n");
    return;
    /// //////////////////

    QVector<double> v_attr(ui.treeAlarms->dataSize(), 0);
    //	QVector<long> v_comm(ui.treeAlarms->dataSize(), 0);
    foreach (QTreeWidgetItem *it, ui.treeAlarms->selectedItems())
    {
        //qDebug() << it->text(3);
        v_attr[it->text(3).toInt()] = 1;
        //		v_comm << it->text(3).toLong();
    }
    try {
        DeviceAttribute attr;
        attr.set_name("Db_alarm");
        vector<double> dati = v_attr.toStdVector();
        attr << dati;
        Tango::DeviceProxy *dpr = DeviceFactory::instance()->getDeviceProxy(modDeviceName);
        if(dpr)
            dpr->write_attribute(attr);
    }
    catch (DevFailed &e)
    {
        TLog log(e);
        time_t now = time(NULL);
        QPair<time_t, DevErrorList> err(now, e.errors);
        TUtil::instance()->addLog(log.toTimeErrlistPair(), log.formatMessage());
        QMessageBox::critical(0, "Error", log.formatMessage());
    }
    ui.treeAlarms->clearSelection();
}

void Interlock::queryConfiguration()
{
    QueryConfiguration win;
    win.setQueryType(queryType);
    if (queryType == QueryConfiguration::Interval)
    {
        win.setStartDateTime(begin);
        win.setStopDateTime(end);
    }

    if (win.exec() == QDialog::Accepted)
    {
        queryType = win.queryType();
        begin = win.startDateTime();
        end = win.stopDateTime();
        readFromDb();
    }
}

void Interlock::readFromDb()
{
    MySqlModel *model = NULL;
    QString dbnam;
    bool modulator = true;
    if(ui.tabWidget->currentIndex() == ModHistory)
    {
        model = modulatorModel;
        dbnam = "real_time_";
    }
    else if(ui.tabWidget->currentIndex() == PsVacHistory)
    {
        modulator = false;
        model = psvacModel;
        dbnam = "vac_mag_";
    }
    else
        return;

    if (!db.isOpen())
    {
        if (!db.open())
        {
            QMessageBox::critical(0, "Error", "Can't establish connection with db FastInterlock on host log");
            return;
        }
    }
    QDateTime adesso = QDateTime::currentDateTime();
    int year = adesso.date().year();

    QString query, endquery;
    if(modulator)
    {
        query = "SELECT FROM_UNIXTIME(" + dbnam + "%1.timestamp) AS 'Date Time', mod_description.name AS 'Name', "
                + dbnam + "%1.event AS 'Event', "
                "mod_description.comment AS Description, mod_description.modulator AS 'Modulator' FROM "
                + dbnam + "%1 left join mod_description on (" + dbnam +
                "%1.id_mod_description=mod_description.id_mod_description) WHERE timestamp>%2 "
                ;
        endquery = " ORDER BY timestamp desc";
    }
    else
    {
        query = "SELECT FROM_UNIXTIME(" + dbnam + "%1.timestamp) AS 'Date Time', vac_mag_desc.name AS 'Name', "
                + dbnam + "%1.event AS 'Event', "
                "vac_mag_desc.comment AS Description, vac_mag_desc.plc  AS 'PLC' FROM "
                + dbnam + "%1 left join vac_mag_desc on (" + dbnam +
                "%1.id_vac_mag_desc=vac_mag_desc.id_vac_mag_desc) WHERE timestamp>%2 "
                ;
        endquery = " ORDER BY timestamp desc";
    }
    unsigned int inizioRicerca;
    switch (queryType)
    {
    case QueryConfiguration::LastDay:
        inizioRicerca = adesso.addDays(-1).toTime_t();
        year = adesso.date().year();
        query = QString(query).arg(year).arg(inizioRicerca) + endquery;
        break;
    case QueryConfiguration::LastWeek:
        inizioRicerca = adesso.addDays(-7).toTime_t();
        year = adesso.date().year();
        query = QString(query).arg(year).arg(inizioRicerca) + endquery;
        break;
    case QueryConfiguration::LastMonth:
        inizioRicerca = adesso.addMonths(-1).toTime_t();
        year = adesso.date().year();
        query = QString(query).arg(year).arg(inizioRicerca) + endquery;
        break;
    case QueryConfiguration::Interval:
        year = end.date().year();
        query = QString(query).arg(year).arg(begin.toTime_t());
        query += QString(" AND timestamp<%1").arg(end.toTime_t()) + endquery;
        break;
    default:
        break;
    }
    model->setQuery(query);

    /* must apply filter if set */
    if(modulator)
        ui.treeLogs->updateFilter();
    else
        ui.treeLogsPSVAC->updateFilter();

    QSqlError error = model->lastError();
    if (error.type() != QSqlError::NoError)
    {
        QMessageBox::critical(0, "Database Error", QString("Error reading from Db:\n") + error.text() + "\nreconnecting");
        db.close();
        if (!db.open())
            QMessageBox::critical(0, "Error", "Can't establish connection with db FastInterlock on host log");
    }
}

void Interlock::on_tabWidget_currentChanged(int idx)
{
    if (idx == 3)
    {
        readFromDb();
        dbTimer->start();
    }
    else
        dbTimer->stop();
}

void Interlock::modSelected(bool en)
{
    //   ui.cbStatSelection->setEnabled(en);
    //   changeStatView(ui.cbStatSelection->currentIndex()); /* will also update descriptions */
}

void Interlock::changeStatView(int index)
{
    QString attname;
    QString devName = modDeviceName;
    /* if Select is checked, then set Stat_mod_NUMBER, otherwise Stat_system */
    //   if(ui.rbSelectMod->isChecked())
    //    attname = QString("Stat_mod_%1").arg(index);
    //   else
    attname = "Stat_system";

    //   QString src = devName + "/" + attname;
    //   ui.tableStates->setSource(src);
}

